#This script illustrates the notion of overfitting

#rm deletes existing variables. 'list' is an argument, and ls() the function to get all existing variables
rm(list=ls())
#ask to press for a key before displaying a new plot
par(ask=TRUE)

#ls.pol is a function that, given a training set (X.tr,Y.tr), builds a polynomial model of degree m, and returns values predicted for the test set X.ts as well as the squarred errors
#Both returned values are stored on a list with argument 'pred' for the model predictions and 'Remp' for the squarred errors
ls.pol<-function(X.tr,Y.tr,m,X.ts){
 #X and X2 contain X.tr and X.ts, respectively,  from degree 1 to m
 X<-NULL
 X2<-NULL
 for (ord in 1:m){
   X<-cbind(X,X.tr^ord)
   X2<-cbind(X2,X.ts^ord)
 }

 #Create the linear model predicting Y given X
 DN<-data.frame(cbind(Y.tr,X))
 names(DN)[1]<-"Y"
 mod<-lm(Y~.,DN)

 #Create the test data frame 
 D.ts<-data.frame(X2)
 names(D.ts)<-names(DN[2:NCOL(DN)])

 #return a list containing:
 # -> argument 'pred', model predictions on the test data frame
 # -> argument 'Remp', mean square error on the test set
 list(pred=predict(mod,D.ts),Remp=mean(mod$residuals^2))

}


#Create a training set for the noisy function y=f(x) on [-1;0.5]
X.tr<-seq(-1,0.5,by=0.05)
Y.tr<-sin(2*pi*X.tr)+rnorm(length(X.tr),sd=0.2)

#Create the test set
X.ts<-seq(-1,0.5,by=0.01)
Y.ts<-sin(2*pi*X.ts)


N<-length(X.tr)

#Create a model with increasing complexity using the training set, and evaluating with the test set.
#m is the order of the polynomial approximation
for (m in 1:(N-2)) {

  #Create the model with function ls.pol, see above
  ls<-ls.pol(X.tr,Y.tr,m,X.ts)

  #Extract predictions and Remp components
  Y.hat<-ls$pred
  Remp<-ls$Remp

  #FPE is a better estimate of the prediction error, taking into account the number of parameters and the number of samples in the observation set.
  FPE<-((1+(m+1)/N)/(1-(m+1)/N))*Remp
 
  #Plot the training set, the test set, and the predicted values for the test set
  plot(X.tr,Y.tr)
  lines(X.ts,Y.ts,lty="dotted")
  lines(X.ts,Y.hat)
  title(paste("Order=",m, "Empirical risk=",round(Remp,digits=4),"FPE=",round(FPE,4)))
  

}
